/*This source code copyrighted by Lazy Foo' Productions 2004-2024
and may not be redistributed without written permission.*/

//Using SDL, strings, and string streams
#include <SDL3/SDL.h>
#include <SDL3/SDL_main.h>
#include <string>
#include <sstream>

//Screen dimension constants
const int SCREEN_WIDTH = 640;
const int SCREEN_HEIGHT = 480;

class LWindow
{
	public:
		//Intializes internals
		LWindow();

		//Creates window
		bool init();

		//Handles window events
		void handleEvent( SDL_Event& e );

		//Focuses on window
		void focus();

		//Shows windows contents
		void render();

		//Deallocates internals
		void free();

		//Window dimensions
		int getWidth();
		int getHeight();

		//Window focii
		bool hasMouseFocus();
		bool hasKeyboardFocus();
		bool isMinimized();
		bool isShown();

	private:
		//Window data
		SDL_Window* mWindow;
		SDL_Renderer* mRenderer;
		int mWindowID;
		int mWindowDisplayIndex;

		//Window dimensions
		int mWidth;
		int mHeight;

		//Window focus
		bool mMouseFocus;
		bool mKeyboardFocus;
		bool mFullScreen;
		bool mMinimized;
		bool mShown;
};

//Starts up SDL and creates window
bool init();

//Frees media and shuts down SDL
void close();

//Our custom window
LWindow gWindow;

//Display data
SDL_DisplayID* gDisplays = NULL;
int gTotalDisplays = 0;
SDL_Rect* gDisplayBounds = NULL; 

int getWindowDisplayIndex( SDL_Window* window )
{
	SDL_DisplayID displayID = SDL_GetDisplayForWindow( window );

	//For simplicity we don't handle display hotplugging, just return the index of the displays enumerated at startup
	for( int i = 0; i < gTotalDisplays; ++i )
	{
		if( displayID == gDisplays[i] )
		{
			return i;
		}
	}

	//Couldn't find the display, was it hotplugged? Just return the main display
	return 0;
}

LWindow::LWindow()
{
	//Initialize non-existant window
	mWindow = NULL;
	mRenderer = NULL;

	mMouseFocus = false;
	mKeyboardFocus = false;
	mFullScreen = false;
	mShown = false;
	mWindowID = -1;
	mWindowDisplayIndex = -1;
	
	mWidth = 0;
	mHeight = 0;
}

bool LWindow::init()
{
	//Create window and renderer
	if( !SDL_CreateWindowAndRenderer( "SDL Tutorial", SCREEN_WIDTH, SCREEN_HEIGHT, SDL_WINDOW_RESIZABLE, &mWindow, &mRenderer ) )
	{
		SDL_Log( "Window could not be created! SDL Error: %s\n", SDL_GetError() );
		return false;
	}

	mMouseFocus = false;
	mKeyboardFocus = false;
	mWidth = SCREEN_WIDTH;
	mHeight = SCREEN_HEIGHT;

	//Enable vsync
	SDL_SetRenderVSync( mRenderer, 1 );

	//Initialize renderer color
	SDL_SetRenderDrawColor( mRenderer, 0xFF, 0xFF, 0xFF, 0xFF );

	//Grab window identifiers
	mWindowID = SDL_GetWindowID( mWindow );
	mWindowDisplayIndex = getWindowDisplayIndex( mWindow );

	//Flag as opened
	mShown = true;

	return true;
}

void LWindow::handleEvent( SDL_Event& e )
{
	//Caption update flag
	bool updateCaption = false;

	//If an event was detected for this window
	if( e.type >= SDL_EVENT_WINDOW_FIRST && e.type <= SDL_EVENT_WINDOW_LAST && e.window.windowID == mWindowID )
	{
		switch( e.type )
		{
			//Window moved
			case SDL_EVENT_WINDOW_MOVED:
			mWindowDisplayIndex = getWindowDisplayIndex( mWindow );
			updateCaption = true;
			break;

			//Window appeared
			case SDL_EVENT_WINDOW_SHOWN:
			mShown = true;
			break;

			//Window disappeared
			case SDL_EVENT_WINDOW_HIDDEN:
			mShown = false;
			break;

			//Get new dimensions and repaint
			case SDL_EVENT_WINDOW_RESIZED:
			mWidth = e.window.data1;
			mHeight = e.window.data2;
			SDL_RenderPresent( mRenderer );
			break;

			//Repaint on expose
			case SDL_EVENT_WINDOW_EXPOSED:
			SDL_RenderPresent( mRenderer );
			break;

			//Mouse enter
			case SDL_EVENT_WINDOW_MOUSE_ENTER:
			mMouseFocus = true;
			updateCaption = true;
			break;
			
			//Mouse exit
			case SDL_EVENT_WINDOW_MOUSE_LEAVE:
			mMouseFocus = false;
			updateCaption = true;
			break;

			//Keyboard focus gained
			case SDL_EVENT_WINDOW_FOCUS_GAINED:
			mKeyboardFocus = true;
			updateCaption = true;
			break;
			
			//Keyboard focus lost
			case SDL_EVENT_WINDOW_FOCUS_LOST:
			mKeyboardFocus = false;
			updateCaption = true;
			break;

			//Window minimized
			case SDL_EVENT_WINDOW_MINIMIZED:
            mMinimized = true;
            break;

			//Window maximized
			case SDL_EVENT_WINDOW_MAXIMIZED:
			mMinimized = false;
            break;
			
			//Window restored
			case SDL_EVENT_WINDOW_RESTORED:
			mMinimized = false;
            break;

			//Hide on close
			case SDL_EVENT_WINDOW_CLOSE_REQUESTED:
			SDL_HideWindow( mWindow );
			break;
		}
	}
	else if( e.type == SDL_EVENT_KEY_DOWN )
	{
		//Display change flag
		bool switchDisplay = false;

		//Cycle through displays on up/down
		switch( e.key.key )
		{
			case SDLK_UP:
			++mWindowDisplayIndex;
			switchDisplay = true;
			break;

			case SDLK_DOWN:
			--mWindowDisplayIndex;
			switchDisplay = true;
			break;
		}

		//Display needs to be updated
		if( switchDisplay )
		{
			//Bound display index
			if( mWindowDisplayIndex < 0 )
			{
				mWindowDisplayIndex = gTotalDisplays - 1;
			}
			else if( mWindowDisplayIndex >= gTotalDisplays )
			{
				mWindowDisplayIndex = 0;
			}

			//Move window to center of next display
			if( !SDL_SetWindowPosition( mWindow, gDisplayBounds[ mWindowDisplayIndex ].x + ( gDisplayBounds[ mWindowDisplayIndex ].w - mWidth ) / 2, gDisplayBounds[ mWindowDisplayIndex ].y + ( gDisplayBounds[ mWindowDisplayIndex ].h - mHeight ) / 2 ) )
			{
				SDL_Log( "Couldn't move window: %s\n", SDL_GetError() );
			}
			updateCaption = true;
		}
	}

	//Update window caption with new data
	if( updateCaption )
	{
		std::stringstream caption;
		caption << "SDL Tutorial - ID: " << mWindowID << " Display: " << mWindowDisplayIndex << " MouseFocus:" << ( ( mMouseFocus ) ? "On" : "Off" ) << " KeyboardFocus:" << ( ( mKeyboardFocus ) ? "On" : "Off" );
		SDL_SetWindowTitle( mWindow, caption.str().c_str() );
	}
}

void LWindow::focus()
{
	//Restore window if needed
	if( !mShown )
	{
		SDL_ShowWindow( mWindow );
	}

	//Move window forward
	SDL_RaiseWindow( mWindow );
}

void LWindow::render()
{
	if( !mMinimized )
	{	
		//Clear screen
		SDL_SetRenderDrawColor( mRenderer, 0xFF, 0xFF, 0xFF, 0xFF );
		SDL_RenderClear( mRenderer );

		//Update screen
		SDL_RenderPresent( mRenderer );
	}
}

void LWindow::free()
{
	if( mWindow != NULL )
	{
		SDL_DestroyWindow( mWindow );
	}

	mMouseFocus = false;
	mKeyboardFocus = false;
	mWidth = 0;
	mHeight = 0;
}

int LWindow::getWidth()
{
	return mWidth;
}

int LWindow::getHeight()
{
	return mHeight;
}

bool LWindow::hasMouseFocus()
{
	return mMouseFocus;
}

bool LWindow::hasKeyboardFocus()
{
	return mKeyboardFocus;
}

bool LWindow::isMinimized()
{
	return mMinimized;
}

bool LWindow::isShown()
{
	return mShown;
}

bool init()
{
	//Initialization flag
	bool success = true;

	//Initialize SDL
	if( SDL_Init( SDL_INIT_VIDEO ) < 0 )
	{
		SDL_Log( "SDL could not initialize! SDL Error: %s\n", SDL_GetError() );
		success = false;
	}
	else
	{
		//Get number of displays
		gDisplays = SDL_GetDisplays( &gTotalDisplays );
		if( gTotalDisplays < 2 )
		{
			SDL_Log( "Warning: Only one display connected!" );
		}
		
		//Get bounds of each display
		gDisplayBounds = new SDL_Rect[ gTotalDisplays ];
		for( int i = 0; i < gTotalDisplays; ++i )
		{
			SDL_GetDisplayBounds( gDisplays[ i ], &gDisplayBounds[ i ] );
		}

		//Create window
		if( !gWindow.init() )
		{
			SDL_Log( "Window could not be created!\n" );
			success = false;
		}
	}

	return success;
}

void close()
{
	//Destroy window
	gWindow.free();

	//Deallocate displays
	SDL_free( gDisplays );
	gDisplays = NULL;
	delete[] gDisplayBounds;
	gDisplayBounds = NULL;

	//Quit SDL subsystems
	SDL_Quit();
}

int main( int argc, char* args[] )
{
	//Start up SDL and create window
	if( !init() )
	{
		SDL_Log( "Failed to initialize!\n" );
	}
	else
	{
		//Main loop flag
		bool quit = false;

		//Event handler
		SDL_Event e;

		//While application is running
		while( !quit )
		{
			//Handle events on queue
			while( SDL_PollEvent( &e ) != 0 )
			{
				//User requests quit
				if( e.type == SDL_EVENT_QUIT )
				{
					quit = true;
				}

				//Handle window events
				gWindow.handleEvent( e );
			}

			//Update window
			gWindow.render();
		}
	}

	//Free resources and close SDL
	close();

	return 0;
}
